import base64
import tempfile
from datetime import datetime
import os
from contextlib import contextmanager
from pathlib import Path

from paramiko import SSHClient, AutoAddPolicy
from scp import SCPClient


@contextmanager
def scp(ssh_client: SSHClient):
    yield SCPClient(ssh_client.get_transport())


@contextmanager
def ssh(server):
    client = SSHClient()
    client.load_system_host_keys()
    client.set_missing_host_key_policy(AutoAddPolicy())
    client.connect(
        server.host,
        username=server.user,
        key_filename=server.key_filename,
        look_for_keys=True,
        timeout=5000
    )
    try:
        yield client
    finally:
        client.close()


@contextmanager
def ssh_key(key, pub_key, key_fname):
    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_dir = Path(tmpdir)
        key_file = tmp_dir.joinpath(key_fname)
        with key_file.open('wb') as wh:
            wh.write(base64.b64decode(key.encode('utf8')))

        pubkey_file = tmp_dir.joinpath(key_fname + '.pub')
        with pubkey_file.open('wb') as wh:
            wh.write(base64.b64decode(pub_key.encode('utf8')))

        try:
            yield str(key_file)
        finally:
            pass


class RemoteServer:
    def __init__(self, host: str, user: str, key_filename: str):
        self.host = host
        self.user = user
        self.key_filename = key_filename

    def get(self, *args, **kwargs):
        with ssh(self) as client:
            with scp(client) as cp:
                cp.get(*args, **kwargs)

    def put(self, *args, **kwargs):
        with ssh(self) as client:
            with scp(client) as cp:
                cp.put(*args, **kwargs)

    @staticmethod
    def tstamp():
        return datetime.utcnow().strftime('%Y%m%d.%H%M%S')


@contextmanager
def ssh_cfg(**kwargs):
    """
    Generating temporary SSH config and dumps ssh keys in temp dir

    Parameters
    ----------
    host : str, 'localhost'
        SSH host, by default is None. If none trying to read env parameter `SSH_HOST`
    user : str, 'guest'
        SSH user, by default is None. If none trying to read env parameter `SSH_USER`
    pkey : str
        Private key, encoded by Base64, by default is None. If none trying to read env parameter `SSH_KEY`
    pub_key : str
        Public key, encoded by Base64, by default is None. If none trying to read env parameter `SSH_PUB_KEY`
    key_fname : str, 'id_rsa'
        Key filename, by default is `id_rsa`.

    Yields
    ------
    conf_path : str
        Temporary SSH Config file location
    """
    ssh_user = kwargs.get('user') or os.getenv('SSH_USER', 'guest')
    ssh_host = kwargs.get('host') or os.getenv('SSH_HOST', 'localhost')
    ssh_alias = kwargs.get('alias') or ssh_host
    ssh_pkey = base64.b64decode(kwargs.get('pkey') or os.getenv('SSH_KEY'))
    ssh_pub_key = base64.b64decode(kwargs.get('pub_key') or os.getenv('SSH_PUB_KEY'))
    ssh_key_fname = kwargs.get('key_fname') or os.getenv('SSH_KEY_FILE', 'id_rsa')

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir_ = Path(tmpdir)

        pkey_path = tmpdir_.joinpath(ssh_key_fname)
        pubkey_path = tmpdir_.joinpath(ssh_key_fname + '.pub')
        conf_path = tmpdir_.joinpath('config')

        with pkey_path.open('wb') as wh:
            wh.write(ssh_pkey)

        with pubkey_path.open('wb') as wh:
            wh.write(ssh_pub_key)

        with conf_path.open('w') as wh:
            wh.write(f'''
Host {ssh_alias}
    IdentitiesOnly  yes
    AddKeysToAgent  yes
    HostName        {ssh_host}
    User            {ssh_user}
    IdentityFile    {pkey_path}
''')

        pkey_path.chmod(0o600)
        pubkey_path.chmod(0o600)
        conf_path.chmod(0o600)

        yield str(conf_path)



import abc
import io
import logging
import multiprocessing as mp
from pathlib import Path

from PIL import Image, ImageOps
from tqdm import tqdm


class CompressorAbc(abc.ABC):
    @abc.abstractmethod
    def params(self) -> dict:
        """Save with compression settings"""


class ImageConverter:
    transparent_color = 255, 255, 255

    def __init__(self, img: Image):
        self.img = img

    @classmethod
    def from_bytes(cls, raw: bytes) -> "ImageConverter":
        return cls(Image.open(io.BytesIO(raw)))

    def scale(self, min_size: int, inplace: bool = False, resample: int = Image.LANCZOS) -> "ImageConverter":
        img = self.img
        if min(img.size) > min_size:
            img = ImageOps.scale(img, min_size / min(img.size), resample=resample)
        if inplace:
            self.img = img
            return self
        return self.__class__(img)

    def to_rgb(self, inplace: bool = False) -> "ImageConverter":
        img = self.img
        if img.mode in ("RGBA", "LA"):
            background = Image.new("RGB", img.size, self.transparent_color)
            background.paste(img, mask=img.split()[-1])
            img = background
        else:
            # img.mode == "P":
            img = self.img.convert("RGB")

        if inplace:
            self.img = img
            return self
        return self.__class__(img)

    def to_bytes(self, compressor: CompressorAbc) -> bytes:
        raw = io.BytesIO()
        self.img.save(raw, **compressor.params())
        return raw.getvalue()

    def to_file(self, file: str or io.BufferedIOBase, compressor: CompressorAbc):
        self.img.save(file, **compressor.params())


class JpegCompressor(CompressorAbc):

    def __init__(self, **kwargs):
        self.format = kwargs.get('format', "JPEG")
        self.dpi = kwargs.get("dpi", (72, 72))
        self.progression = kwargs.get("progression", True)
        self.quality = kwargs.get("quality", 70)
        self.optimize = kwargs.get("optimize", True)
        self.progressive = kwargs.get("progressive", True)

    def params(self) -> dict:
        return dict(
            format=self.format,
            dpi=self.dpi,
            progression=self.progression,
            quality=self.quality,
            optimize=self.optimize,
            progressive=self.progressive
        )


def compare_images(img_file: Path, min_size: int = 700, quality: int = 70) -> tuple:
    try:
        jpeg = JpegCompressor(quality=quality)
        initial_length = Path(img_file).stat().st_size

        im = Image.open(io.BytesIO(img_file.read_bytes()))
        print(f'MODE: {im.mode}')
        # print(f'{im.split()}')

        ib = ImageConverter(im) \
            .scale(min_size=min_size, inplace=True) \
            .to_rgb(inplace=True) \
            .to_bytes(compressor=jpeg)

        return initial_length, len(ib)
    except Exception as exc:
        logging.error(f'{img_file}: {exc}')


def scan_folder(folder: Path):
    for obj in folder.iterdir():
        if obj.is_dir():
            continue
        if obj.name in ('.DS_Store', ):
            continue

        old_size, new_size = compare_images(obj)
        compression = (old_size - new_size) / old_size * 100
        print(f"{obj.relative_to(folder)} {old_size} --> {new_size} / {compression:.2f}%")

scan_folder(Path('/Users/ilya/Downloads/ttttttt'))


def convert_images(img_file: Path, min_size: int = 700, quality: int = 70) -> Path:
    try:
        jpeg = JpegCompressor(quality=quality)
        with Image.open(img_file) as img:
            imb = ImageConverter(img)

        imb.scale(min_size=min_size, inplace=True) \
            .to_rgb(inplace=True) \
            .to_file(img_file, compressor=jpeg)

        return img_file
    except Exception as exc:
        raise Exception(f"{img_file} error {exc}") from exc


def convert_images_master(logfile: str, base_path: Path, processes=None):
    with mp.Pool(processes=processes) as pool:
        with open(logfile, 'r') as files:
            fn_args = (
                base_path.joinpath(file.strip())
                for file in files
            )
            it = pool.imap_unordered(convert_images, fn_args)
            with open(logfile + '.progress', 'w') as log:
                for result in tqdm(it, desc='Converting images'):
                    log.write(str(result) + "\n")



import operator as op

import pytest

from grabutils.bs.pageobj import (
    BsField, BsPageObj, PageObject, Href, Image, Nested,
    inner_text, stripped, as_attr, List
)


class PageObjectConstant(PageObject):
    hrefs = BsField('a', as_attr('href'), many=True)
    labels = BsField('a', inner_text, many=True)
    divs = BsField('div', inner_text, many=True)
    constant1 = 1

    @staticmethod
    def attr_fname(*args, **kwargs):
        return 'fname'

    @property
    def constant2(self):
        return 2

    class Meta:
        # fields = 'href', 'label', 'constant1', 'constant2'
        ignore_fields = 'divs',


class Language1(BsPageObj):
    href = BsField('a', as_attr('href'), many=True)
    label = BsField('a', inner_text, many=True)


class Language2(BsPageObj):
    href = BsField(None, as_attr('href'))
    label = BsField(None, inner_text)


class Language3(BsPageObj):
    languages = BsField('a', Nested(Language2), many=True)


class PageObject(BsPageObj):
    paragraph = BsField('p', inner_text, stripped)
    link = Href('.test-url')
    image = Image('.test-img')

    items = BsField('ul > li', inner_text, stripped, many=True)
    languages = BsField('.languages a', inner_text, many=True)
    languages_1 = BsField('.languages', Nested(Language1))
    languages_2 = BsField('.languages a', Nested(Language2), many=True)
    languages_3 = BsField('.languages > div', List(BsField('a', Nested(Language2), many=True)), many=True)

    _invisible = BsField('.languages a', inner_text, many=True)

    @property
    def custom_field(self):
        return [
            v for idx, v in enumerate(self._invisible)
            if idx in (1, 2, 4)
        ]


@pytest.fixture(scope="module")
def page(html):
    return PageObject(html)


def test_page_object(html):
    pageobj = PageObjectConstant(html)
    value = pageobj.as_dict()

    assert value == {
        'constant1': 1,
        'constant2': 2,
        'fname': 'fname',
        'hrefs': ['https://www.python.org/',
                  'http://ada',
                  'https://java',
                  'http://cpp',
                  'http://cobol',
                  'http://d',
                  'http://go'],
        'labels': ['Python', 'Ada', 'Java', 'C++', 'Cobol', 'D', 'Go']
    }


def test_link(page):
    link = page.link

    assert 'href' in link
    assert link['href'] == 'https://www.python.org/'

    assert 'label' in link
    assert link['label'] == 'Python'


def test_paragraph(page):
    assert (page.paragraph == "Hello, world!")


def test_image(page):
    image = page.image

    assert 'src' in image
    assert image['src'] == 'https://www.python.org/static/img/python-logo@2x.png'

    assert 'label' in image
    assert image['label'] == 'Python logo'


def test_items(page: PageObject):
    items = page.items

    assert items == ['Alpha', 'Beta', 'Gamma']


def test_nested(page: PageObject):
    languages = page.languages
    languages_1 = page.languages_1
    languages_2 = page.languages_2
    languages_3 = page.languages_3

    assert languages == ['Ada', 'Java', 'C++', 'Cobol', 'D', 'Go']
    assert languages_1 == {
        'href': ['http://ada', 'https://java', 'http://cpp', 'http://cobol', 'http://d', 'http://go'],
        'label': ['Ada', 'Java', 'C++', 'Cobol', 'D', 'Go']
    }
    assert languages_2 == [
        {'href': 'http://ada', 'label': 'Ada'},
        {'href': 'https://java', 'label': 'Java'},
        {'href': 'http://cpp', 'label': 'C++'},
        {'href': 'http://cobol', 'label': 'Cobol'},
        {'href': 'http://d', 'label': 'D'},
        {'href': 'http://go', 'label': 'Go'}
    ]
    assert languages_3 == [
        [{'href': 'http://ada', 'label': 'Ada'}, {'href': 'https://java', 'label': 'Java'}],
        [{'href': 'http://cpp', 'label': 'C++'}],
        [{'href': 'http://cobol', 'label': 'Cobol'}],
        [{'href': 'http://d', 'label': 'D'}, {'href': 'http://go', 'label': 'Go'}]
    ]
    assert page.custom_field == [
        'Java', 'C++', 'D'
    ]

    assert '_invisible' not in page.as_dict()
    assert 'custom_field' in page.as_dict()



import functools as fn
import operator as op

from bs4 import BeautifulSoup, Tag


class BsField:
    def __init__(self, selector, *then, src_attr='page', many=False):
        self.src_attr = src_attr
        self.selector = selector
        self.many = many
        self.then = then

    def __call__(self, node):
        then = self.then

        if self.many:
            res = node.select(self.selector) if self.selector else node
            return [
                fn.reduce(lambda v, func: func(v), then, r)
                for r in res
                if r
            ] if then else res

        else:
            res = node.select_one(self.selector) if self.selector else node
            return fn.reduce(lambda v, func: func(v), then, res) if then and res else res

    def __get__(self, obj, owner):
        if not (obj and isinstance(obj, BsPageObj) or isinstance(obj, PageObject)):
            return self

        if isinstance(obj, BsPageObj):
            page = obj.page()
        else:
            page = obj.as_source()
        then = self.then

        return self.__call__(page)


class PageObjMeta(type):
    def __new__(mcs, name, bases, attrs):
        klass = super().__new__(mcs, name, bases, attrs)

        fields = set(
            field_name
            for field_name, field
            in attrs.items()
            if (
                isinstance(field, BsField) or
                (callable(field) and field_name.startswith('attr_')) or
                not (
                    field_name.startswith('_') or
                    field_name == 'Meta' or
                    callable(field)
                )

            )
        )

        meta = klass.Meta

        meta_fields = getattr(meta, 'fields') if hasattr(meta, 'fields') else None
        meta_ignore_fields = getattr(meta, 'ignore_fields') if hasattr(meta, 'ignore_fields') else None
        meta_parser = getattr(meta, 'parser') if hasattr(meta, 'parser') else 'lxml'

        if meta_fields and meta_ignore_fields:
            raise Exception('Both attributes *fields* and *ignore_fields* are declared in Meta class ')

        if meta_fields:
            fields &= set(meta_fields)

        if meta_ignore_fields:
            fields -= set(meta_ignore_fields)

        klass._parser = meta_parser
        klass._export_fields = tuple(fields)
        return klass


class PageObject(metaclass=PageObjMeta):

    def __init__(self, page: str):
        if isinstance(page, BeautifulSoup) or isinstance(page, Tag):
            self.__page = page
        else:
            self.__page = BeautifulSoup(str(page), features=self._parser)
        self.__value = None

    def as_source(self):
        return self.__page

    def __extract_key(self, field):
        attr = getattr(self, field)
        if callable(attr):
            return field.replace('attr_', ''), attr(self.__page)

        return field, attr

    def as_dict(self):
        if not self.__value:
            self.__value = dict(
                self.__extract_key(field)
                for field
                in self._export_fields
            )
        return self.__value

    class Meta:
        parser: str = 'lxml'
        fields: tuple = None
        ignore_fields: tuple = None


class BsPageObj:

    @staticmethod
    def parser() -> str:
        return 'lxml'

    @staticmethod
    def system_fields() -> tuple:
        return 'parser', 'as_dict', 'page'

    @staticmethod
    def ignore_fields() -> tuple:
        return tuple()

    @classmethod
    def transient_attrs(cls) -> set:
        return set(cls.system_fields() + cls.ignore_fields())

    def __init__(self, page: str):
        if isinstance(page, BeautifulSoup) or isinstance(page, Tag):
            self.__page = page
        else:
            self.__page = BeautifulSoup(str(page), features=self.parser())
        self.__class_fields = None

    def page(self):
        return self.__page

    def as_dict(self):
        # list of BsFields (from class)
        class_fields = set(
            fld
            for fld, inst
            in self.__class__.__dict__.items()
            if isinstance(inst, BsField)
        )

        transient_attrs = self.transient_attrs()

        return dict(
            (field, getattr(self, field))
            for field
            in dir(self)
            if
            (
                field in class_fields or
                field not in transient_attrs

            ) and
            not field.startswith('_') and
            not field[0].isupper() and
            not callable(getattr(self, field))
        )


class Nested:
    def __init__(self, page_obj_clazz):
        self.page_obj_clazz = page_obj_clazz

    def __call__(self, value):
        return self.page_obj_clazz(value).as_dict()


inner_text = op.attrgetter('string')


def stripped(value: str):
    return value.strip() if value else ''


def as_attr(attr, default=None):
    def as_attr_wrapper(node):
        return node.attrs.get(attr, default)

    return as_attr_wrapper


def parse_href(node):
    return dict(
        href=node.attrs['href'],
        label=node.string.strip()
    )


def parse_img(node):
    return dict(
        src=node.attrs['src'],
        label=node.attrs.get('alt') if node else None
    )


def prepend(text: str):
    def wrapper(value: str):
        return text + value if isinstance(value, str) else value

    return wrapper


class Href(BsField):
    def __init__(self, selector, *then, **kwargs):
        super(Href, self).__init__(selector, parse_href, *then, **kwargs)


class Image(BsField):
    def __init__(self, selector, **kwargs):
        super(Image, self).__init__(selector, parse_img, **kwargs)


class List:
    def __init__(self, field: BsField):
        self.field = field

    def __call__(self, node):
        return self.field.__call__(node)



from subprocess import Popen, PIPE
import logging


def shexec(cmd, stdout=True) -> int or None:
    proc = Popen(
        cmd,
        stderr=PIPE,
        stdout=PIPE,
        shell=True,
        universal_newlines=True
    )

    logging.debug(cmd)

    if stdout:
        for line in proc.stdout:
            print(line.strip())

    errors = False
    for line in proc.stderr:
        if not errors:
            logging.error(cmd)
            errors = True
        logging.error(line)

    return proc.returncode



import re

import pytest

from grabutils.dictutils import dictview


@pytest.fixture
def data() -> dictview:
    return dictview({
        'a': {'b': 1},
        'b': [{'c': 'c'}, 'd', 4, ('e', 5)],
        'c': ('d', 6),
        'd': {'e', 'f', 'g', 'h'},
        'e': None
    })


def test_value(data: dictview):
    # a = data.a
    assert data.a.value == {'b': 1}
    assert data.b[0].value == {'c': 'c'}
    assert data.b[100].value is None
    assert data.e.value is None
    assert data.not_existing_key.value is None
    assert data['not_existing_key'].value is None
    assert data[1000].value is None


def test_value_ctx(data: dictview):
    with data.a.value_ctx as value:
        assert value == {'b': 1}

    with data.b[0].value_ctx as value:
        assert value == {'c': 'c'}

    with data.b[100].value_ctx as value:
        assert value is None

    with data.e.value_ctx as value:
        assert value is None

    with data.not_existing_key.value_ctx as value:
        assert value is None

    with data['not_existing_key'].value_ctx as value:
        assert value is None

    with data.not_existing_key[1000].value_ctx as value:
        assert value is None

    with data[1000].value_ctx as value:
        assert value is None


def test_scan(data):
    assert data.scan('a').value == {'b': 1}
    assert data.scan('b.0').value == {'c': 'c'}
    assert data.scan('b.3.1').value == 5
    assert data.scan('b.100').value is None
    assert data.scan('not_existing_key').value is None
    assert data.scan('a.b.c.d.e.f.-1000').value is None


def test_regex(data):
    pattern = re.compile('^(a|b|e|f)$')
    assert data[pattern].value == [
        {'b': 1},
        [{'c': 'c'}, 'd', 4, ('e', 5)],
        None
    ]



import re
from contextlib import contextmanager


class dictview:

    def __init__(self, data, inline=False):
        self.__container_data = data
        self.inline = inline

    @property
    def value(self):
        return self.__container_data

    # @value.setter
    # def value(self, value):
    #     self.__container_data = value

    @property
    @contextmanager
    def value_ctx(self):
        yield self.__container_data

    def scan(self, path: str) -> 'dictview':
        """
        Scan

        Parameters
        ----------
        path : str
            specific path in current dictview, i.e. "path.to.the.content"

        Returns
        -------
        dictview
        """

        d = dictview(self.value, inline=True)
        for el in path.split('.'):
            if re.match(r'^\d+$', el):
                d = d[int(el)]
            else:
                d = d[el]
        d.inline = self.inline
        return d

    def project(self, **kwargs):
        return dict(
            (k, self.scan(path).value)
            for k, path in kwargs.items()
        )

    def __str__(self):
        return str(self.__container_data)

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        if hasattr(self.__container_data, '__iter__'):
            return self.__container_data.__iter__()

    def __len__(self):
        try:
            return self.__container_data.__len__()
        except:
            return 0

    def __get_value(self, item):
        if not self.__container_data:
            ret = self.__container_data
        elif hasattr(self.__container_data, item):
            ret = getattr(self.__container_data, item)
        elif hasattr(self.__container_data, '__getitem__') and isinstance(item, int):
            ret = self.__container_data.__getitem__(item)
        elif hasattr(self.__container_data, 'get'):
            ret = self.__container_data.get(item)
        elif isinstance(self.__container_data, set):
            ret = item if item in self.__container_data else None
        else:
            ret = None

        return self._wrap_output(ret)

    def __search(self, pattern: re.Pattern) -> list:
        if isinstance(self.__container_data, dict):
            return [v for k, v in self.__container_data.items() if pattern.search(k)]
        elif isinstance(self.__container_data, list) or isinstance(self.__container_data, tuple):
            return [v for v in self.__container_data if pattern.search(v)]
        elif isinstance(self.__container_data, set):
            return [v for v in self.__container_data if pattern.search(v)]
        else:
            return []

    def _wrap_output(self, output):
        if self.inline:
            self.__container_data = output
            return self

        return dictview(output, inline=self.inline)

    def __getitem__(self, item) -> 'dictview':
        if self.__container_data is None:
            return self._wrap_output(self.__container_data)

        if hasattr(self.__container_data, '__getitem__'):
            try:
                if isinstance(item, re.Pattern):
                    res = self.__search(item)
                else:
                    res = self.__container_data.__getitem__(item)
            except IndexError:
                res = None
            except KeyError:
                res = None

            return self._wrap_output(res)

        if item == '__container_data':
            return super(dictview, self).__getattr__(item)

        try:
            res = getattr(self.__container_data, item)
        except AttributeError:
            res = None

        return self._wrap_output(res)

    def __contains__(self, item) -> bool:
        if hasattr(self.__container_data, '__contains__'):
            return self.__container_data.__contains__(item)
        elif isinstance(item, str):
            return hasattr(self.__container_data, item)
        else:
            return False

    def __getattr__(self, item) -> 'dictview':
        assert isinstance(item, str)
        res = None
        finished = False

        if item == '__container_data':
            return super(dictview, self).__getattr__(item)

        if hasattr(self.__container_data, '__getitem__'):
            try:
                res = self.__container_data.__getitem__(item)
                finished = True
            except IndexError:
                finished = True
            except KeyError:
                finished = False

        if not finished:
            try:
                res = getattr(self.__container_data, item)
                finished = True
            except AttributeError:
                finished = False

        return self._wrap_output(res)



from functools import reduce


class Project:
    def __init__(self, *path, **kwargs):
        self.path = path
        self.field = kwargs.get('field')
        self.silent = kwargs.get('silent', False)
        self.default = kwargs.get('default')
        self.conv = kwargs.get('astype')
        # self.allow_none = kwargs.get('allow_none', True)

    @staticmethod
    def get_field(obj, field, silent, default) -> (bool, any):
        if hasattr(obj, '__getitem__'):
            try:
                return False, obj[field]
            except KeyError as ke:
                if not silent:
                    raise ke
                return True, default
        elif isinstance(field, str) and hasattr(obj, field):
            return False, getattr(obj, field)

    def iter_path(self, obj, field):
        is_done, old_obj = obj
        if not is_done:
            return self.get_field(old_obj, field, self.silent, self.default)
        return obj

    def __call__(self, obj):
        was_failed, content = reduce(self.iter_path, self.path, (False, obj))
        return self.conv(content) if self.conv else content

    def __get__(self, obj, owner) -> any:
        if hasattr(obj, '_get_data'):
            return self(getattr(obj, '_get_data')())
        else:
            return self


class Zip(Project):

    def __init__(self, left: Project, right: Project, **kwargs):
        super(Zip, self).__init__(*[left, right], **kwargs)
        self.conv = self.conv or list

    def __call__(self, obj):
        return self.conv(zip(*[el(obj) for el in self.path]))


class Callable(Project):

    def __init__(self, fn, **kwargs):
        super(Callable, self).__init__(fn, **kwargs)

    def __call__(self, obj):
        content = self.path[0](obj)

        return self.conv(content) if self.conv else content

    def __get__(self, obj, owner) -> any:
        if hasattr(obj, '_get_data'):
            return self(getattr(obj, '_get_data')())
        else:
            return self




class ObjectViewMeta(type):
    def __new__(mcs, name, bases, attrs):
        klass = super().__new__(mcs, name, bases, attrs)

        dynamic_fields = set()

        for field_name, field in attrs.items():
            if isinstance(field, Project):
                field.field = field_name
            elif callable(field) and field_name.startswith('attr_'):
                dynamic_fields.add(field_name.replace('attr_', ''))

        fields = set(
            field_name
            for field_name, field
            in attrs.items()
            if (
                isinstance(field, Project) or
                (callable(field) and field_name.startswith('attr_')) or
                not (
                    field_name.startswith('_') or
                    field_name == 'Meta' or
                    callable(field)
                )

            )
        )
        meta = klass.Meta

        meta_fields = getattr(meta, 'fields') if hasattr(meta, 'fields') else None
        meta_ignore_fields = getattr(meta, 'ignore_fields') if hasattr(meta, 'ignore_fields') else None

        if meta_fields and meta_ignore_fields:
            raise Exception('Both attributes *fields* and *ignore_fields* are declared in Meta class ')

        if meta_fields:
            fields &= set(meta_fields)

        if meta_ignore_fields:
            fields -= set(meta_ignore_fields)

        klass._export_fields = tuple(fields)
        klass._dynamic_fields = dynamic_fields
        return klass


class ObjectView(metaclass=ObjectViewMeta):
    def __init__(self, data):
        self.__data = data
        # for fld, fn in self._dynamic_fields.items():
        #     setattr(self, fld, Callable(fn))

    def _get_data(self):
        return self.__data

    def __getattr__(self, item):
        if item in self._dynamic_fields:
            return getattr(self, 'attr_' + item)(self.__data)
        return super(ObjectView, self).__getattribute__(item)

    class Meta:
        fields = []
        ignore_fields = []



from functools import reduce, partial

import jmespath


class JsReader:
    def __init__(self, value: any):
        self.jsview = value


class JsPath:
    def __init__(self, path: str):
        self.js_path = path
        self._init = False
        self._value = None
        self._proc = list()

    def map(self, fn):
        def mmap(value):
            return fn(value)

        self._proc.append(mmap)
        return self

    def flatmap(self, fn):
        def mflatmap(value):
            return map(fn, value)

        self._proc.append(mflatmap)
        return self

    def filter(self, jmespath_query: str):
        self._proc.append(partial(jmespath.search, jmespath_query))
        return self


    def __get__(self, obj: JsReader, objtype=None):
        if self._init:
            return self._value
        self._value = reduce(lambda memo, item: item(memo), self._proc, jmespath.search(self.js_path, obj.jsview))
        return self._value



import os


class TextFileUtils:

    @staticmethod
    def partition(file: str, partitions: int = 0) -> list:
        """Partition text file into partitions
        Parameters
        ----------
        file : str
            file to partition
        partitions : int
            num partitions, by default equal `os.cpu_count()`

        Returns
        -------
        list
            List of created partitions

        """
        if int(partitions) <= 0:
            partitions = os.cpu_count()

        files = [
            f'{file}.{partition:04d}'
            for partition in range(partitions)
        ]

        with open(file, 'r') as lines:
            for line_no, line in enumerate(lines, start=1):
                partition = line_no % partitions
                with open(files[partition], 'a') as file_partition:
                    file_partition.write(file)

        return files



import time
import sys
import math
import logging

# Code was taken from IPython
# Taken from here: https://github.cosm/ipython/ipython/blob/master/IPython/core/magics/execution.py

try:
    import resource


    def clock():
        """clock() -> floating point number
        Return the *TOTAL USER+SYSTEM* CPU time in seconds since the start of
        the process.  This is done via a call to resource.getrusage, so it
        avoids the wraparound problems in time.clock()."""

        u, s = resource.getrusage(resource.RUSAGE_SELF)[:2]
        return u + s


    def clock2():
        """clock2() -> (t_user,t_system)
        Similar to clock(), but return a tuple of user/system times."""
        return resource.getrusage(resource.RUSAGE_SELF)[:2]
except ImportError:
    # There is no distinction of user/system time under windows, so we just use
    # time.perff_counter() for everything...
    clock = time.perf_counter


    def clock2():
        """Under windows, system CPU time can't be measured.
        This just returns perf_counter() and zero."""
        return time.perf_counter(), 0.0


class Timer:
    logger = logging.getLogger(__name__)

    @classmethod
    def wrap(cls, fn):
        def wrapper(*args, **kwargs):
            with Timer():
                res = fn(*args, **kwargs)
            return res

        return wrapper

    @staticmethod
    def wall_time():
        return time.time()

    @staticmethod
    def fmt(timespan, precision=3):
        """Formats the timespan in a human readable form"""
        if timespan >= 60.0:
            # we have more than a minute, format that in a human readable form
            # Idea from http://snipplr.com/view/5713/
            parts = [("d", 60 * 60 * 24), ("h", 60 * 60), ("min", 60), ("s", 1)]
            times = []
            leftover = timespan
            for suffix, length in parts:
                value = int(leftover / length)
                if value > 0:
                    leftover = leftover % length
                    times.append(u'%s%s' % (str(value), suffix))
                if leftover < 1:
                    break
            return " ".join(times)

        # Unfortunately the unicode 'micro' symbol can cause problems in
        # certain terminals.
        # See bug: https://bugs.launchpad.net/ipython/+bug/348466
        # Try to prevent crashes by being more secure than it needs to
        # E.g. eclipse is able to print a µ, but has no sys.stdout.encoding set.
        units = [u"s", u"ms", u'us', "ns"]  # the save value
        if hasattr(sys.stdout, 'encoding') and sys.stdout.encoding:
            try:
                u'\xb5'.encode(sys.stdout.encoding)
                units = [u"s", u"ms", u'\xb5s', "ns"]
            except:
                pass
        scaling = [1, 1e3, 1e6, 1e9]

        if timespan > 0.0:
            order = min(-int(math.floor(math.log10(timespan)) // 3), 3)
        else:
            order = 3
        return u"%.*g %s" % (precision, timespan * scaling[order], units[order])

    def __repr__(self):
        ret = ""

        if sys.platform != 'win32':
            cpu_user = self.fmt(self.cpu_user)
            cpu_sys = self.fmt(self.cpu_sys)
            cpu_tot = self.fmt(self.cpu_tot)
            ret += f"CPU times: user {cpu_user}, sys: {cpu_sys}, total: {cpu_tot}\n"

        wall_time = self.fmt(self.wall_time)
        ret += f"Wall time: {wall_time}\n"
        return ret

    def __init__(self):
        self.wall_start = None
        self.wall_end = None
        self.cpu_start = None
        self.cpu_end = None

        self.wall_time = None
        self.cpu_user = None
        self.cpu_sys = None
        self.cpu_tot = None

    def __enter__(self):
        self.wall_start = self.wall_time()
        self.cpu_start = clock2()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.wall_end = self.wall_time()
        self.cpu_end = clock2()

        self.wall_time = self.wall_end - self.wall_start
        self.cpu_user = self.cpu_end[0] - self.cpu_start[0]
        self.cpu_sys = self.cpu_end[1] - self.cpu_start[1]
        self.cpu_tot = self.cpu_user + self.cpu_sys

        self.logger.debug(repr(self))



import typing as t

import pytest

from grabutils.objview import Project, ObjectView, Zip


@pytest.fixture
def data1() -> dict:
    return {
        'current': {
            'field': ['is', 'too', {'nested': True}],
            'fields': ['a', 'b', 'c'],
            'values': [1, 2, 3],
            'notfoo': 'notfoo',
            'nested': [
                {'currency': 'USD', 'amount': 600.},
                {'currency': 'AUD', 'amount': 1200.},
                {'currency': 'EUR', 'amount': 30.},
            ]
        },
        'data': {
            'aaaaaa': {
                'content': 'Hello, World'
            },
            'dddddd': {
                'content': 'Shalom!'
            },
            'bbbbb': {
                '__ref': 'aaaaaa'
            },
            'cccccc': {
                '__refs': ['bbbbb', 'dddddd']
            }
        }
    }


@pytest.fixture
def model_class1() -> t.Type[ObjectView]:
    class Ref(Project):
        """
        Ref substitutor
        """

        def __init__(self, ref_path: t.Iterable[any], *path, **kwargs):
            super(Ref, self).__init__(*path, **kwargs)
            self.ref_path = ref_path

    class MyView(ObjectView):
        is_nested = Project('current', 'field', 2, 'nested')
        is_too = Project('current', 'field', 1)
        is_keyerror = Project('current', 'fake')
        is_keyerror2 = Project('current', 0)

        transpose = Zip(
            Project('current', 'fields'),
            Project('current', 'values'),
        )
        transpose1 = Zip(
            Project('current', 'fields'),
            Project('current', 'values'),
            astype=dict
        )
        constant = 4

        def attr_foo(self, obj):
            return obj['current']['notfoo']

    return MyView


def test_objview1(data1: dict, model_class1: t.Type[ObjectView]):
    model = model_class1(data1)

    assert isinstance(model.is_nested, bool)
    assert model.is_nested is True
    assert model.is_too == 'too'

    with pytest.raises(KeyError):
        assert model.is_keyerror is None

    with pytest.raises(KeyError):
        assert model.is_keyerror2 == 10

    assert model.transpose == [
        ('a', 1),
        ('b', 2),
        ('c', 3),
    ]

    assert model.transpose1 == {
        'a': 1,
        'b': 2,
        'c': 3,
    }

    assert model.constant == 4

    assert model.foo == 'notfoo'



import logging
from abc import ABC

from urllib.parse import urljoin, urlencode, quote_plus

import requests


class Field:
    pass


class Dto(ABC):
    _fields = tuple()

    @classmethod
    def __is_field(cls, f: str) -> bool:
        clz = getattr(cls, f).__class__
        return issubclass(clz, Field)

    @classmethod
    def init_dto(cls):
        cls._fields = tuple(f for f in dir(cls) if cls.__is_field(f)) + tuple([cls.key_field()])
        cls.__slot__ = cls._fields + tuple([
            '__modified',

        ])

    @classmethod
    def key_field(cls) -> str:
        return "id"

    def key_exists(self) -> bool:
        return hasattr(self, self.key_field())

    def get_key(self):
        return getattr(self, self.key_field())

    def __init__(self, data: dict):
        super().__init__()
        if data:
            self.update(**data)

    def items(self, full=False) -> iter:
        for f in self._fields:
            if f in self._fields:
                if not hasattr(self, f):
                    if full:
                        yield f, None
                    else:
                        pass
                else:
                    yield f, getattr(self, f)

    def to_dict(self) -> dict:
        return dict(self.items())

    def update(self, **data):
        for f in data:
            if f in self._fields:
                setattr(self, f, data[f])


class Endpoint:
    logger = logging.getLogger(__name__)

    def __init__(self, base, **params):
        self.base = base
        self.params = params

    def __repr__(self):
        if not self.params:
            return self.base
        return self.base + "?" + urlencode(self.params, quote_via=quote_plus)

    def __call__(self, *context, **params):
        paramz = dict(self.params)
        paramz.update(**params)
        return Endpoint(urljoin(self.base + '/', '/'.join(map(quote_plus, map(str, context))), **paramz))

    def __getattr__(self, key: str):
        if key in {'get', 'post', 'put', 'delete'}:
            logger = self.logger
            url = self.base

            def wrap(*args, **kwargs):
                logger.debug(f"{key.upper()} {url}")
                method = getattr(requests, key)
                return method(url, *args, **kwargs)

            return wrap
        return Endpoint(urljoin(self.base + '/', key), **self.params)



import abc
import io

from PIL import Image, ImageOps


class CompressorAbc(abc.ABC):
    @abc.abstractmethod
    def params(self) -> dict:
        """Save with compression settings"""


class ImageConverter:
    transparent_color = 255, 255, 255

    def __init__(self, img: Image):
        self.img = img

    @classmethod
    def from_bytes(cls, raw: bytes) -> "ImageConverter":
        return cls(Image.open(io.BytesIO(raw)))

    def scale(self, min_size: int, inplace: bool = False, resample: int = Image.LANCZOS) -> "ImageConverter":
        img = self.img
        if min(img.size) > min_size:
            img = ImageOps.scale(img, min_size / min(img.size), resample=resample)
        if inplace:
            self.img = img
            return self
        return self.__class__(img)

    def to_rgb(self, inplace: bool = False) -> "ImageConverter":
        img = self.img
        if img.mode in ("RGBA", "LA"):
            background = Image.new("RGB", img.size, self.transparent_color)
            background.paste(img, mask=img.split()[-1])
            img = background
        else:
            # img.mode == "P":
            img = self.img.convert("RGB")

        if inplace:
            self.img = img
            return self
        return self.__class__(img)

    def to_bytes(self, compressor: CompressorAbc) -> bytes:
        raw = io.BytesIO()
        self.img.save(raw, **compressor.params())
        return raw.getvalue()

    def to_file(self, file: str or io.BufferedIOBase, compressor: CompressorAbc):
        self.img.save(file, **compressor.params())


class JpegCompressor(CompressorAbc):

    def __init__(self, **kwargs):
        self.format = kwargs.get('format', "JPEG")
        self.dpi = kwargs.get("dpi", (72, 72))
        self.progression = kwargs.get("progression", True)
        self.quality = kwargs.get("quality", 70)
        self.optimize = kwargs.get("optimize", True)
        self.progressive = kwargs.get("progressive", True)

    def params(self) -> dict:
        return dict(
            format=self.format,
            dpi=self.dpi,
            progression=self.progression,
            quality=self.quality,
            optimize=self.optimize,
            progressive=self.progressive
        )



import setuptools

with open("readme.md", "r") as fh:
    long_description = fh.read()

bs_deps = [
    'beautifulsoup4',
    'lxml',
]

jspo_deps = [
    'jmespath'
]

ssh_deps = [
    'paramiko[ed25519]',
    'scp',
]
rest_deps = [
    'requests'
]

img_deps = [
    'Pillow'
]

setuptools.setup(
    name="grabutils",
    version="0.0.31",
    author="Ilya Babich",
    author_email="sniter@gmail.com",
    description="Utils and helpers for crawling websites",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sniter/grabutils",
    install_requires=[],
    extras_require={
        'bs': bs_deps,
        'jspo': jspo_deps,
        'ssh': ssh_deps,
        'rest': rest_deps,
        'img': img_deps,
        'all': bs_deps + ssh_deps + rest_deps + img_deps + jspo_deps
    },
    tests_require=[
        'pytest'
    ],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



def str_to_bool(value):
    if not value:
        return False
    if isinstance(value, bool):
        return value
    if not isinstance(value, str):
        value = str(value)
    return value.lower() in ('yes', 'true', '1')
